#!/usr/bin/perl -w

#   Copyright (C) 2006 David G. Jacobowitz 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#


#perl2exe_include "bytes.pm"


BEGIN {
 # find and load all 'loadable' perl modules.
 use FindBin;
 use lib $FindBin::Bin;
 require find_command_modules;
};
our @registerable_modules;

use strict;
use matlibctrl   qw(use_pari use_pdl);
use misc         qw(banner);
use cmd_driver   qw(handleCommandFile handleCommand);
use Time::HiRes  qw(gettimeofday);
use Cwd;
use Data::Dumper;
use gui          qw(setupGui setGuiCmdsRef tk_MainLoop register_menus);

use logging;
push (@registerable_modules, "logging");

$| = 1;


#### globals 
my $dm = {};               # data model
my $gb = { conf => 0.95,
	   collinearity_threshold => 0.999,
	   last_time  => [gettimeofday],
	   start_time => [gettimeofday],
	   using_gui => 1,
           using_pdl     => use_pdl(),
           using_parilib => use_pari(),
	   helpfile      => "atats_help.dat",
           dirty => 0,
           version_major => "0",
           version_minor => "11",
           build_number => "303",
           build_date => "Tue Jun 19 21:45:00 2012",
	   installdir => $FindBin::Bin,
	   platform   => $^O,
	   head_lines => 10,
           graphics_formats => ["emf", "png", "postscript", "eps", "bmp", "gif"] };


my $oo = logging->new();

my $stdout      = *STDOUT;
my $stderr      = undef;
my $save_err    = 0;
my $cmdfname    = undef;
my $use_cmdfile = 0;
my $showver     = 0;

##################################
# process command line arguments
##################################
while (@ARGV) {
 my $linearg = shift;
 if ($linearg =~ /^-f/) {
  $cmdfname = shift;
  if (defined($cmdfname) && (-e $cmdfname)) {
   $use_cmdfile = 1;
  } else {
   $oo->disp("Error. Command file $cmdfname not found.\n");
   exit(1);
  }
 } elsif ($linearg =~ /^-g/) {
  $gb->{using_gui} = !$gb->{using_gui};
 } elsif ($linearg =~ /^--?v/) {
  $showver = 1;
 } elsif ($linearg =~ /^-stdin/) {
  $gb->{using_gui} = 0;
 } elsif ($linearg =~ /^-e/) {
  $save_err = 1;
 } else {
 };
}

if ($showver) {
 print STDOUT banner($gb);
 exit(0);
}

if ($gb->{using_gui}) {
 setupGui($oo,$gb,$dm);
 register_menus($gb->{tktree});
};


##################################
# build command list
##################################
# Each object has some methods that are to be callable
# directly by the user in ATATS; the objects identify
# these with a list returned by "->commands()". The
# remaining methods are to be used in Perl-land only --
# not in AT-ATS command land.
my $modules  = {};
my $commands = {};
foreach my $rm (@registerable_modules) {
 if ($rm eq "logging") {
  $modules->{$rm}{instance} = $oo;
 } else {
  $modules->{$rm}{instance} = $rm->new($oo,$gb,$dm);
 }
 my @new_cmds = $modules->{$rm}{instance}->commands();
 foreach my $cmd (@new_cmds) {
  $commands->{$cmd} = $modules->{$rm}{instance};
 }
}

if ($gb->{using_gui}) {
 setGuiCmdsRef($commands);
}

###################################
# print out startup status messages
###################################
$oo->tdisp('banner',banner($gb));

if ($save_err) {
 my $errfname = "err_atats.log";
 $oo->tdisp('brite'," + Saving debugging log file to $errfname.\n");
 open(ERRFILE,"> " . $errfname);
 $stderr = *ERRFILE;
 $oo->set_stderr($stderr);
};

if (use_pdl()) {
 $oo->tdisp('brite'," + Using Perl::PDL matrix routines. ($PDL::VERSION)\n");
} elsif (use_pari()) {
 $oo->tdisp('brite'," + Using libPari matrix routines. (fast but buggy)\n");
} else {
 $oo->tdisp('brite'," . Using Math::MatrixReal matrix routines (pure perl).\n");
}

$oo->tdisp('brite'," + Started on: " . localtime($gb->{start_time}->[0]) . ".\n");

if ($gb->{have_gnuplot}) {
 $oo->tdisp('brite'," + GNUplot detected. Graphing enabled.\n");
} else {
 $oo->tdisp('brite'," - GNUplot not detected. No graphing..\n");
}

$oo->tdisp('brite'," . Current directory: " . Cwd::cwd() . "\n");
$oo->disp("\n");

##################################
# If directory has initialization file, run that.
##################################
my $inifile = ".atats";
if (-e $inifile) {
 $oo->disp("Processing commands from user initialization file...\n\n");
 handleCommandFile($commands,$oo,$gb,$dm,$inifile);
}

##################################
# start processing commands
##################################

$oo->disp("\n");


if ($gb->{using_gui}) {
 if ($use_cmdfile) {
  handleCommandFile($commands,$oo,$gb,$dm,$cmdfname);
 }
 tk_MainLoop();
}  else {
 if ($use_cmdfile) {
  handleCommandFile($commands,$oo,$gb,$dm,$cmdfname);
 } else {
  while (1) {
   $oo->cdisp(". ");
   my $cmd = <STDIN>;
   chomp $cmd;
   handleCommand($commands,$oo,$gb,$dm,$cmd);
  }
 }
}



$oo->tdisp('banner',"\n Goodbye!\n\n");
$oo->done();
exit(0);
