#!/usr/bin/perl -w

#   Copyright (C) 2006 David G. Jacobowitz 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

use fgroup;
package fgroup::gen;
BEGIN { @ISA = qw(fgroup); };

use Exporter 'import';
@EXPORT_OK = qw();

use strict;
use misc qw(step min max xtile randnorm);
use misc2 qw(inAry inDict);

use Scalar::Util qw(looks_like_number);
use eqn_string qw(create_evaluable_eqn);
use gui qw(tk_updateVars);
use safe_eval qw(safe_eval);
use Data::Dumper;

sub commands {
 my $self = shift;
 my @cmds = ("gen", "generate", "replace");
 return @cmds;
}

sub replace {
 my $self   = shift;
 my $a      = shift;
 my $oo     = $self->{oo};

 my $r      = {};

 my $set_eqn = "";

 my $newvar = "";
 if (!$a->{firstvar}{present} || ($a->{firstvar}{state} ne "exist")) {
  $r->{status} = "fail; replace variable $newvar does not exist";
  return $r;
 } else {
  $newvar = $a->{firstvar}{name};
 }

 if (!($a->{have_if} && $a->{if_ok})) {
  return { status => "fail; missing \"if\" expression" }
 }

 if ($a->{have_eval}) {
  $set_eqn = $a->{eval_str};
 } else {
  return { status => "fail; no assignment expression" }
 }

 my $set_eval_str = "";
 $r = create_evaluable_eqn($self->{dm},$self->{gb},$set_eqn,0);
 if (!($r->{status} eq "ok")) {
  return $r;
 } else {
  $set_eval_str      = $r->{expression};
  $r->{set_eval_str} = $r->{expression};
 }

 my $row = 0;

 my @tary    = @{ $self->{dm}{data}{$newvar} };
 my $rows    = $#tary + 1;
 my $numeric = $self->{dm}{info}{$newvar}{type};
 my $pcount  = 0;

 for (my $row=0;$row<$rows;$row++) {
  my $lrow = undef;
  if (defined($a->{reverse_iftable}{$row})) { 
   $lrow = $a->{reverse_iftable}{$row}; 
  };
  my ($newval, $problem, $report) = safe_eval($set_eval_str,
	                                      $self,
					      $row,
					      $lrow,
				              $a->{forward_iftable});
  my $ifval = $a->{ifary}[$row];
  # print "newval $newval ifval $ifval\n";
  if ($ifval) {
   $self->{dm}{data}{$newvar}[$row] = $newval;
   if ($problem) { 
    $pcount++; 
    $oo->debug($report . "\n");
   };
   if (!looks_like_number($newval)) {
    $numeric = "categorical";
   }
  }
 }
 $self->{dm}{info}{$newvar}{type} = $numeric;
 $r->{status} = "ok";
 if ($pcount) { $oo->disp("Warning: $pcount undefs generated.\n\n"); };
 return $r;
}



sub gen {
 my $self  = shift;
 my $a     = shift;
 my $oo     = $self->{oo};
 my $r     = { status => "ok" };

 my $newvar = "";
 if (!$a->{firstvar}{present}) {
  $r->{status} = "fail; new variable not specified";
  return $r;
 } else {
  $newvar = $a->{firstvar}{name};
 }

 my $set_eqn;
 if ($a->{have_eval}) {
  $set_eqn = $a->{eval_str};
 } else {
  return { status => "fail; no assignment expression" }
 }

 $r = create_evaluable_eqn($self->{dm},$self->{gb},$set_eqn);
 if (!($r->{status} eq "ok")) {
  return $r;
 }

 my $eval_str = $r->{expression};
 my @vars = keys %{$self->{dm}{data}};
 my $sample_var = shift(@vars);

 my $rows = 0;
 if (defined($sample_var)) {
  my @tary = @{ $self->{dm}{data}{$sample_var} };
  $rows = $#tary + 1;
 } else {
  if (defined($self->{dm}{locals}{obscount})) {
   $rows = $self->{dm}{locals}{obscount};
  } else {
   return { status => "fail; you are calling generate on an empty dataset. Load some data first, or else, set the local variable \"obscount\" to define the number of observations to create." };
  }
 }

 # do we need to clear out the variable to undef first if it
 # does not exist -- yes, if there is also an 'if' array 
 # present
 if (!defined($self->{dm}{data}{$newvar})) {
  if (defined($a->{ifary})) {
   for (my $i=0;$i<$rows;$i++) { 
    $self->{dm}{data}{$newvar}[$i] = "__undef";
   }
  }
 }

 my $numeric = "numeric";
 my $pcount    = 0;
 my $skipcount = 0;
 for (my $row=0;$row<$rows;$row++) {
  my $lrow = undef;
  if (defined($a->{reverse_iftable}{$row})) { 
   $lrow = $a->{reverse_iftable}{$row}; 
  };

  my ($newval, $problem, $report) = safe_eval($eval_str,
	                                      $self,
					      $row,
					      $lrow,
				              $a->{forward_iftable});

  if ($problem) {
   $pcount++;
   $oo->debug($report . "\n");
  }

  if ($newval eq "") { 
   $newval = 0;
  } 
  if (($newval ne "__undef") && !looks_like_number($newval)) {
   $numeric = "categorial";
  };
  
  if (defined($a->{ifary}[$row])) {
   if ($a->{ifary}[$row]) {
    $self->{dm}{data}{$newvar}[$row] = $newval;
   } else {
    $skipcount++;
   }
  } else {
   $self->{dm}{data}{$newvar}[$row] = $newval;
  }
 }
 $self->{dm}{info}{$newvar}{type} = $numeric;

 if ($self->{gb}{using_gui}) {
  tk_updateVars($self->{dm});
 }

 $r->{status} = "ok";
 if ($pcount)    { $oo->disp("warn; $pcount undefs generated\n"); };
 if ($skipcount) { $oo->disp("warn; $skipcount rows unassigned\n"); };
 if ($pcount || $skipcount) { $oo->disp("\n"); }
 $r->{pcount} = $pcount;
 $r->{skipcount} = $skipcount;
 $r->{numeric}   = $numeric;
 return $r;
}


sub generate {
 my $self = shift;
 my $a    = shift;
 return $self->gen($a);
}

1;

