#!/usr/bin/perl -w

#   Copyright (C) 2006 David G. Jacobowitz 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

use fgroup;
package fgroup::generics;
BEGIN { @ISA = qw(fgroup); };

use Exporter 'import';
@EXPORT_OK = qw();

use warnings 'all';
use strict;

use misc2 qw(inAry inDict setFlg tstFlg func_fail);
use eqn_string qw(create_evaluable_eqn create_evaluable_eqn_immed);
use gui qw(tk_updateVars);
use Cwd;
use Scalar::Util qw(looks_like_number);
use Data::Dumper;
use matrix_help qw(set_matrix_mem);
use Time::HiRes qw(gettimeofday);
use dataqual qw(gen_ifquald_a);
use safe_eval qw(safe_eval);
use YAML      qw(DumpFile);


sub commands {
 my $self = shift;
 my @cmds = ("done", "quit", "q", "qc", "exit", "set", "clear", "rval", "gval", "drop", "level", "sortt", "vars", "count", "list", "return");
 return @cmds;
}

sub done {
 my $self  = shift;
 my $a     = shift;
 my $oo    = $self->{oo};

 my $force = 0;
 if (tstFlg($a,"clear")) {
  $force = 1;
 }
 my $gb     = $self->{gb};

 if (!$force && $gb->{dirty}) {
  $oo->disp("> fail; no. data in memory would be lost\n");
  return { status => "ok" };
 } else {
  return { status => "shutdown" };
 }
}

sub quit {
 my $self = shift;
 return $self->done(@_);
}

sub qc {
 my $self = shift;
 my $a    = shift;
 setFlg($a,"clear",1);
 return $self->done($a,@_);
}

sub q {
 my $self = shift;
 return $self->done(@_);
}

sub exit {
 my $self = shift;
 return $self->done(@_);
}

sub clear {
 my $self    = shift;
 $self->{gb}{dirty} = 0;
 delete $self->{dm}{data};
 delete $self->{dm}{info};
 delete $self->{gb}{rval};
 delete $self->{gb}{rval2};
 delete $self->{last_run};
 tk_updateVars($self->{dm});
 return { status => "ok" };
}

sub set {
 my $self    = shift;
 my $a       = shift;
 my $oo     = $self->{oo};
 my $gb     = $self->{gb};
 my $dm     = $self->{dm};

 my $varname = "";
 if ($a->{firstvar}{present}) {
  if ($a->{firstvar}{name} eq "mem") {
   if ($gb->{using_parilib}) {
    my $newval = $a->{args}[1];
    $oo->disp("> setting mem to $newval\n");
    set_matrix_mem($newval);
    return { status => "ok" };
   } else {
    return { status => "ok; whatever" };
   }
  }
  $varname = $a->{firstvar}{name};
 } else {
  foreach my $v ( keys %{$gb} ) {
   if (defined($v) && 
       defined($gb->{$v})) {
    $oo->disp("   set " . $v . " " . $gb->{$v} . "\n");
   }
  }
  $oo->disp("\n\n");
  return { status => "ok" };
 }

 my $eqn_str = "";
 if ($a->{have_eval}) {
  $eqn_str = $a->{eval_str};
 } else {
  return { status => "fail; no assignment specified" }
 }

 # print "EQN $eqn_str\n";
 my $r = create_evaluable_eqn_immed($self->{dm},$self->{gb},$eqn_str);
 if (func_fail($r)) { return $r; };

 my ($newval, $problem, $report) = safe_eval($r->{expression},$self,0,0);
 if ($problem) {
  $oo->debug($report . "\n");
 }
 if (!defined($newval)) {
  return { status => "fail; new value malformed" }
 } else {
  $oo->disp(" set: $varname = $newval\n");
  $self->{gb}{$varname} = $newval;
  return { status => "ok" };
 }
}

sub level {
 my $self = shift;
 my $a    = shift;
 if (!$a->{firstvar}{present}) {
  return { status => "fail; no value present" }
 }
 my $newval = $a->{firstvar}{name};
 if (!looks_like_number($newval)) {
  return { status => "fail; value not numeric" }
 }
 if (($newval > 1) || ($newval < 0)) {
  return { status => "fail; not in range [0,1]" }
 }
 $self->{gb}{conf} = $newval;
 return { status => "ok" };
}


sub drop {
 my $self   = shift;
 my $a      = shift;
 my $oo     = $self->{oo};
 my $r      = { status => "ok" };

 if (!$a->{firstvar}{present}) {
  # dropping rows
  if (!($a->{have_if} && $a->{if_ok})) {
   return { status => "fail; dropping observations requires an 'if' expression" }
  } else {
   if (0) {
    return { status => "fail; could not generate if expression" }
   } else {
    my @vars = keys %{$self->{dm}{data}};
    if (!@vars) {
     return { status => "fail; empty dataset" };
    }
    my $newrow = 0;
    my $dropcount = 0;
    for (my $i=0;$i<@{$self->{dm}{data}{$vars[0]}};$i++)  {
     if (!$a->{ifary}[$i]) {
      foreach my $var (@vars) {
       $self->{dm}{newdata}{$var}[$newrow] = $self->{dm}{data}{$var}[$i];
      }
      $newrow++;
     } else {
      $dropcount++;
     }
    } 
    $oo->disp("$dropcount observations dropped\n\n");
    delete $self->{dm}{data};
    $self->{dm}{data} = $self->{dm}{newdata};
    delete $self->{dm}{newdata};
   }
  }
 } else {
  # dropping vars
  foreach my $v (keys %{$a->{vars}}) {
   my $e_str = "";
   my $ifval = 0;
   if ($a->{have_if}) {
    $r = create_evaluable_eqn_immed($self->{dm},$self->{gb},$a->{if_str});
    if (func_fail($r)) { return $r; }
    my $lrow = undef;
    my ($ifval, $problem, $report) = safe_eval($r->{expression},$self,0,0);
    if ($problem) {
     $oo->debug($report . "\n");
    }
   }
   $ifval |= !$a->{have_if};
   if ($ifval) {
    if ($a->{vars}{$v} eq "exist") {
     delete $self->{dm}{data}{$v};
     delete $self->{dm}{info}{$v};
    } else {
     $oo->disp(" Could not drop nonexistant \"$v\"\n"); 
    }
   }
  }
 }

 if ($self->{gb}{using_gui}) {
  tk_updateVars($self->{dm});
 };
 return { status => "ok" };
}


sub dictCSV {
 my $d      = shift;
 my $indent = shift;
 my $spaces = "";
 for (my $i=0;$i<$indent;$i++) {
  $spaces .= ',';
 }

 my $os = '';
 foreach my $da (sort keys %{$d}) {
  if (ref($d->{$da}) eq "HASH") {
   $os .= sprintf("%s%s:\n",$spaces,$da);
   $os .= dictCSV($d->{$da},$indent+1);
  } elsif (ref($d->{$da}) eq "ARRAY") { 
   $os .= sprintf("%s%s:\n",$spaces,$da);
   $os .= arryCSV($d->{$da},$indent+1);
  } elsif (ref($d->{$da}) eq "") {
   if (defined($da) && 
       defined($d->{$da}) &&
       defined($spaces)) {
    $os .= sprintf("%s%s,%s\n",$spaces,$da, $d->{$da});
   }
  }
 }
 return $os;
}

sub arryCSV {
 my $d      = shift;
 my $indent = shift;
 my $spaces = "";
 for (my $i=0;$i<$indent;$i++) {
  $spaces .= ',';
 }
 my $i=0;
 my $os = '';
 foreach my $da (@{$d}) {
  if (ref($da) eq "HASH") {
   $os .= sprintf("%s%s:\n",$spaces,$i);
   $os .= dictCSV($da,$indent+1);
  } elsif (ref($da) eq "ARRAY") { 
   $os .= arryCSV($da,$indent+1);
  } elsif (ref($da) eq "") {
   if (defined($da) && 
       defined($spaces)) {
    $os .= sprintf("%s%s,%s\n",$spaces,$i, $da);
   }
  }
 $i++;
 }
 return $os;
}

sub dictShow {
 my $oo     = shift;
 my $d      = shift;
 my $indent = shift;
 my $spaces = "";
 for (my $i=0;$i<$indent;$i++) {
  $spaces .= "  ";
 }
 foreach my $da (sort keys %{$d}) {
  if (ref($d->{$da}) eq "HASH") {
   $oo->disp(sprintf("%s %-20s:\n",$spaces,$da));
   dictShow($oo,$d->{$da},$indent+1);
  } elsif (ref($d->{$da}) eq "ARRAY") { 
   $oo->disp(sprintf("%s %-20s:\n",$spaces,$da));
   arryShow($oo,$d->{$da},$indent+1);
  } elsif (ref($d->{$da}) eq "") {
   if (defined($da) && 
       defined($d->{$da}) &&
       defined($spaces)) {
    $oo->disp(sprintf("%s %-20s = %20s\n",$spaces,$da, $d->{$da}));
   }
  }
 }
}

sub arryShow {
 my $oo     = shift;	
 my $d      = shift;
 my $indent = shift;
 my $spaces = "";
 for (my $i=0;$i<$indent;$i++) {
  $spaces .= "  ";
 }
 my $i=0;
 foreach my $da (@{$d}) {
  if (ref($da) eq "HASH") {
   $oo->disp(sprintf("%s %-20s:\n",$spaces,$i)); #$da));
   dictShow($oo,$da,$indent+1);
  } elsif (ref($da) eq "ARRAY") { 
   arryShow($oo,$da,$indent+1);
  } elsif (ref($da) eq "") {
   if (defined($da) && 
       defined($spaces)) {
    $oo->disp(sprintf("%s %-20s = %20s\n",$spaces,$i, $da));
   }
  }
 $i++;
 }
}

sub return {
 my $self = shift;
 my $oo   = $self->{oo};
 my $a    = shift;
 if (defined($a->{args}[0]) && ($a->{args}[0] eq "list")) {
  $self->rval();
 } else {
  if (defined($self->{gb}{rval}{$a->{args}[0]})) {
   
   $oo->disp($a->{args}[0] . 
	     " = " .  
	     $self->{gb}{rval}{$a->{args}[0]} . 
	     "\n");
  } else {
   $oo->disp("value not found\n");
  }
 }
 return $self->{gb}{rval};
}

sub rval {
 my $self    = shift;
 my $oo      = $self->{oo};
 my $a       = shift;
 $oo->disp("\nReturn Values:\n\n");
 dictShow($oo,$self->{gb}{rval},0);
 $oo->disp("\n");

 if (defined($a->{options}{output})) {
  my $fname = $a->{options}{output}{str};
  if ($fname =~ /\.csv$/i) {
   my $ofh;
   open($ofh,'>',$fname);
   if (defined($ofh)) {
    print $ofh dictCSV($self->{gb}{rval},0);
    close $ofh;
   } else {
    $oo->warn("could not open file $fname");
   }
  } elsif ($fname =~ /\.yaml$/i) {
   DumpFile($fname,$self->{gb}{rval});
  }
 };
 return $self->{gb}{rval};
}

sub gval {
 my $self    = shift;
 my $oo      = $self->{oo};
 $oo->disp("\nReturn Values:\n\n");
 dictShow($oo,$self->{gb},0);
 $oo->disp("\n");
 return $self->{gb}{rval};
}

sub sortt {
 my $self     = shift;
 my $ar       = shift;
 my $r        = { status => "ok" };

 my $svar = "";
 if (!$ar->{firstvar}{present} && ($ar->{firstvar}{state} ne "exist")) {
  return { status => "sort variable not specified or nonexistent" };
 } else {
  $svar = $ar->{firstvar}{name};	  
 }

 my @var = @{$self->{dm}{data}{$svar}};
 my @vvar = ();
 for (my $i=0;$i<@var;$i++) {
  $vvar[$i]->{value} = $var[$i];
  $vvar[$i]->{idx}   = $i;
 }

 my @svvar = sort { (looks_like_number($a->{value}) && looks_like_number($b->{value})) ?  
	            ($a->{value} <=> $b->{value}) : 
		    ($a->{value} cmp $b->{value}) } @vvar;

 my $ndata= {};
 for (my $i=0;$i<@svvar;$i++) {
  foreach my $var (keys %{$self->{dm}{data}}) {
   $ndata->{$var}[$i] = $self->{dm}{data}{$var}[$svvar[$i]->{idx}];
  }
 }
 delete $self->{dm}{data};
 $self->{dm}{data} = $ndata;
 return $r;
};


sub vars {
 my $self = shift;
 my $oo   = $self->{oo};
 my @vars = keys %{$self->{dm}{data}};
 my $r = {};
 $r->{status} = "ok";
 if (@vars) {
  my $os = "Available variables: \n";
  foreach my $v (@vars) {
   $os .= "  $v\n";
  }
  $os .= "\n\n";
  $oo->disp($os);
  $r->{vars} = \@vars;
 } else {
  $oo->disp("No variables defined.\n\n");
 };
 return $r;
}

sub count {
 my $self = shift;
 my $a    = shift;
 my $oo   = $self->{oo};
 my $ct   = 0;
 my @vars = keys %{$self->{dm}{data}};
 my $tot = 0;
 if (defined($self->{dm}{data}{$vars[0]})) {
  $tot = @{$self->{dm}{data}{$vars[0]}};
 }
 if (defined($a->{ifary})) {
  for (my $i=0;$i<$tot;$i++) {
   if ($a->{ifary}[$i]) { $ct++; };
  }
 } else {
  $ct = $tot;
 }
 $oo->disp(" $ct\n\n");
 my $r = {};
 $r->{count}  = $ct;
 $r->{status} = "ok";
 return $r;
}

sub list {
 my $self = shift;
 my $a    = shift;
 my $oo   = $self->{oo};
 my @vars = ();
 foreach my $v (@{$a->{varary}}) {
  if ($v->{state} eq "exist") {
   push (@vars,$v->{name});
  }
 }
 if (!@vars) {
  @vars = sort keys %{$self->{dm}{data}};
 }


 my $qha    = {};
 my $ocount = 0;
 foreach my $v (@vars) {
  my $qa = gen_ifquald_a($self->{dm},$v,$a->{ifary});
  $qha->{$v} = $qa;
  $ocount = @{$qa};
 }

 my $os = "";
 for (my $i=0;$i<$ocount;$i++) {
  $os .= sprintf("%d: -------------------------------------\n",$i+1);
  foreach my $v (@vars) {
   $os .= sprintf("%15s = %10s\n",$v,$qha->{$v}[$i]); 
  }
  $os .= "\n";
 }
 $oo->disp($os);

 return { status => "ok" };
}


1;

