#!/usr/bin/perl -w

#   Copyright (C) 2006 David G. Jacobowitz 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

use fgroup;
package fgroup::ols_lr;
BEGIN { @ISA = qw(fgroup); };

use Exporter 'import';
@EXPORT_OK = qw(regress_pretty_print calc_ols_WaldF);
use warnings 'all';

use Data::Dumper;
use misc qw(covariance fixedFloat fixedFloat_p deep_copy);
use dataqual qw(gen_numquald_ha);

use misc2 qw(inAry inDict tstFlg func_fail);

use eqn_string qw(create_evaluable_eqn create_evaluable_eqn_immed str_to_varlist);
use gui qw(tk_updateVars);
use matrix_ols qw(m_ols_regress m_robust_ols m_calc_ols_wald_s m_2sls_regress);
use matrix_help qw(m_gen_sqrt_diag_a m_return_col_a m_mean_tot_col_s m_ss_col_s a_gen_matrix_from_cols_m a_gen_matrix_from_rows_m m_get_elem_s de_objify gen_zero_matrix_m m_mat_to_string m_ret_size_a m_sum_sq_s m_set_elem);
# use Statistics::Distributions;
use mod_distributions;
use Statistics::Distributions;

use strict;

sub commands {
 my $self = shift;
 my @cmds = ("regress", "reg", "predict", "exclusionrestrict", "test", "vce");
 return @cmds;
}

sub predict {
 my $self      = shift;
 my $a         = shift;
 my $argref    = $a->{args};
 my $flagref   = $a->{flgs};

 my @args      = @{$argref};
 my @flags     = @{$flagref};

 my $newvar = "";
 if (!defined($a->{firstvar})) {
  return { status => "fail; no variable to create specified" };
 } else {
  $newvar = $a->{firstvar}{name};
 }

 if (defined($self->{gb}{last_run}{matrices}{pred_v})) {
  for (my $y=0;$y<$self->{gb}{last_run}{orig_depth};$y++) {
   $self->{dm}{data}{$newvar}[$y] = "__undef";
   # print "setting $newvar [$y] = __undef\n";
  }
  my $data = [];
  if (tstFlg($a,"resid")) {
   $data = m_return_col_a($self->{gb}{last_run}{matrices}{resid_v},1);
  } else {
   $data = m_return_col_a($self->{gb}{last_run}{matrices}{pred_v},1);
  }
  # print Dumper $self->{gb}{last_run};
  for (my $y=0;$y<@{$data};$y++) {
   my $old_row = $self->{gb}{last_run}{orig_from}[$y];
   $self->{dm}{data}{$newvar}[$old_row] = $data->[$y];
#   print "setting $newvar [$old_row] = $data->[$y]\n";
  }

  $self->{dm}{info}{$newvar}{type} = "numeric";

  if ($self->{gb}{using_gui}) {
   tk_updateVars($self->{dm});
  }
  return { status => "ok" };
 } else {
  return { status => "fail; no regression available. Was last command a regression?" };
 }
}

sub pre_regress_collinearity_check {
 my $self = shift;
 my $indr = shift;
 my $qdm  = shift; ## this is the datamodel with undefs and nonnums 
                   ## already removed and "ifs" already applied
 my $r = {};
 $r->{status} = "ok";
 my @indeps = @{$indr};

 foreach my $ev ( @indeps ) {
  if (!defined($qdm->{data}{$ev})) {
   $r->{status} = "fail;  indep var \"$ev\" does not exist";
   return $r;
  }

  foreach my $ev2 ( @indeps ) {
   if (!defined($qdm->{data}{$ev2})) {
    $r->{status} = "fail; indep var \"$ev2\" does not exist";
    return $r;
   }
   if ($ev2 eq $ev) {
   } else {
    my $res = covariance($qdm->{data}{$ev},
	                 $qdm->{data}{$ev2});
    if (func_fail($res)) {
     return $res;
    } elsif (abs($res->{correlation}) > $self->{gb}{collinearity_threshold}) {
     $res->{status} = "fail -- $ev and $ev2 are nearly collinear . Try dropping one from the regression";
     return $res;
    }
   }
  }
 }
 return $r;
}

sub regress_calc {

 my $self      = shift;
 my $a         = shift;
 my $options   = shift;

 my $depvar;

 my $r = { status => "ok" };

 if ($a->{firstvar}{state} eq "exist") {
  $depvar = $a->{firstvar}{name};
 } else {
  return { status => "fail; no dependent variable found" };
 }  

 my @indeps = ();
 foreach my $v (@{$a->{latervars}}) {
  if ($v->{state} eq "exist") {
   push (@indeps, $v->{name});
  } else {
   return { status => "fail; independent variable $v->{name} not found" };
  }
 }

 if (!@indeps) {
  return { status => "fail; no independent variables provided" };
 }

 # possible vars includes things found in the "aweight", if present
 my @possible_vars = ();
 if (defined($a->{options}{aweight})) {
  @possible_vars = str_to_varlist($a->{options}{aweight}{str},
	                            $self->{dm});
 }

 my @allvars = ($depvar, @indeps, @possible_vars);

 my $qdm = {};
 $qdm = gen_numquald_ha($self->{dm},\@allvars, $a->{ifary});

 # this should not be necessary anymore becuase the ols routines
 # first check to make sure the matrices are invertible before
 # trying to invert. Still, we don't say which equations are 
 # collineary; that's a problem for another day.
 # $r = $self->pre_regress_collinearity_check(\@indeps,$qdm);
 # if (func_fail($r)) {
 #  return $r;
 # }

 # store last analysis for certain tests that compare 2
 # regressions
 if (defined($self->{gb}{last_run})) {
  $self->{gb}{last_last_run} = deep_copy($self->{gb}{last_run});
 }

 delete $self->{gb}{last_run};

 # store information about what undef/not-if values were dropped
 # in preparation for the regression run; we'll need this for 
 # the 'predict' command later.
 $self->{gb}{last_run}{orig_from}  = deep_copy($qdm->{orig_from});
 $self->{gb}{last_run}{orig_depth} = $qdm->{orig_depth};

 $r->{depvar}              = $depvar;
 $self->{gb}{last_run}{depvar} = $depvar;


 ### First, populate Arrays to make regression calculations
 ##########################################################


 my $dv_a = [];
 my $ev_aa = [];

 my @ev_ref_ary = ();

 foreach my $ev ( @indeps ) {
  push (@ev_ref_ary, $qdm->{data}{$ev});
 }

 my @ones_array = ();

 my $ev_count  = @indeps;
 my $obs_count = 0;
 if (defined($ev_ref_ary[0])) {
  $obs_count = @{$ev_ref_ary[0]};
 }
 if ($obs_count <= ($ev_count+1)) {
  return { status => "fail; insufficient observations: $obs_count obs, $ev_count indep variables" };
 }

 foreach my $row (@{$ev_ref_ary[0]}) {
  push(@ones_array,1.0);
 }

 if (($a->{have_opt}) && defined(($a->{options}{aweight}))) {
  # print we have to deal with weighting
  my $tr = create_evaluable_eqn($qdm,
	                        $self->{gb},
				$a->{options}{aweight}{str},1);
  if (func_fail($tr)) { return $tr; };

  my $eval_str = $tr->{expression};
  my $weights  = $tr->{ifary};
  my $weighted_dv = [];

  $ev_aa = deep_copy(\@ev_ref_ary);

  for (my $row=0;$row<@{$qdm->{data}{$depvar}};$row++) {
   $weighted_dv->[$row] = $qdm->{data}{$depvar}[$row] * $weights->[$row];
   for (my $i=0;$i<@{$ev_aa};$i++) {
    $ev_aa->[$i][$row] = $ev_aa->[$i][$row] * $weights->[$row];
   }
  } 
  $dv_a  = $weighted_dv;
  $ev_aa = [ \@ones_array, @{$ev_aa} ];
 } else {
  $dv_a  = $qdm->{data}{$depvar};
  $ev_aa = [ \@ones_array, @ev_ref_ary ];
 }
 my $dv_v = a_gen_matrix_from_cols_m([$dv_a]);
 my $ev_m = a_gen_matrix_from_cols_m($ev_aa);
 my $N    = @{$dv_a};
 my $k    = @{$ev_aa} - 1;
 
 ### Next, find coefficients
 ##########################################################

 my $disp_m    = undef;
 my $soln_m    = undef;
 my $resid_v   = undef;
 my $pred_v    = undef;
 my $rss       = undef;
 if (tstFlg($a,"robust")) {
  ($disp_m, $soln_m, $pred_v, $resid_v, $rss) = m_robust_ols($ev_m,$dv_v);
 } else {
  ($disp_m, $soln_m, $pred_v, $resid_v, $rss) = m_ols_regress($ev_m,$dv_v);
 }
 if (!defined($disp_m)) {
  $r->{status} = "fail; could not create dispersion matrix -- collinearity problem?";
  return $r;
 }


 $self->{gb}{last_run}{type}              = tstFlg($a,"robust") ? "robust" : "";
 $self->{gb}{last_run}{matrices}{ev_m}    = $ev_m;
 $self->{gb}{last_run}{matrices}{dv_v}    = $dv_v;
 $self->{gb}{last_run}{matrices}{disp_m}  = $disp_m;
 $self->{gb}{last_run}{matrices}{soln_m}  = $soln_m;
 $self->{gb}{last_run}{matrices}{pred_v}  = $pred_v;
 $self->{gb}{last_run}{matrices}{resid_v} = $resid_v;
 $r->{indeps}                         = \@indeps;
 $self->{gb}{last_run}{indeps}            = \@indeps;
 my $i = 1;
 foreach my $ev ( ("_cons", @indeps) ) {
  my $coeff = m_get_elem_s($soln_m,$i,1);
  $r->{coeffs}{$ev} = $coeff;
  $self->{gb}{last_run}{coeffs}{$ev} = $coeff;
  $i++;
 }

 ### Calculate predicted values and residuals;
 ##########################################################

 (my $actmean, my $acttot) = m_mean_tot_col_s($self->{gb}{last_run}{matrices}{dv_v},1);

# my $rss = m_ss_col_s($resid_v,0,1);
 my $tss = m_ss_col_s($dv_v,$actmean,1);
 my $ess = m_ss_col_s($pred_v,$actmean,1);

 $r->{tss}              = $tss;
 $r->{ess}              = $ess;
 $r->{rss}              = $rss;
 $self->{gb}{last_run}{tss} = $tss;
 $self->{gb}{last_run}{ess} = $ess;
 $self->{gb}{last_run}{rss} = $rss;

 ### Calculate some statistics
 ##########################################################

 my $rsquared = $ess/$tss;
 $r->{rsquared}              = $rsquared;
 $self->{gb}{last_run}{rsquared} = $rsquared;

 my $mst   = $tss / ($N-1);
 my $mse   = $ess / ($k);
 my $msr   = $rss / ($N-($k+1));
 $r->{mst} = $mst;
 $r->{mse} = $mse;
 $r->{msr} = $msr;
 $self->{gb}{last_run}{mst} = $mst;
 $self->{gb}{last_run}{mse} = $mse;
 $self->{gb}{last_run}{msr} = $msr;

 $r->{k}   = $k;
 $r->{N}   = $N;
 $self->{gb}{last_run}{k}   = $k;
 $self->{gb}{last_run}{N}   = $N;

 my $t_crit = Statistics::Distributions::tdistr(($N-($k+1)),
	                                        ((1-$self->{gb}{conf})/2));
 $r->{t_crit}              = $t_crit;
 $self->{gb}{last_run}{t_crit} = $t_crit;

 my $sigma_e    = sqrt($msr);
 $r->{sigma_e}                 = $sigma_e;
 $self->{gb}{last_run}{sigma_e}    = $sigma_e;


 if (!tstFlg($a,"robust")) {

  # !!!! this is broken.... how does one do this.
  my $adj_rsquared = $rsquared * ($N-($k+1)) / ($N-1);
  $r->{adj_rsquared}              = $adj_rsquared;
  $self->{gb}{last_run}{adj_rsquared} = $adj_rsquared;

  # !!!! this is probably wrong, too
  my $F = $mse / $msr;
  my $F_prob = abs(mod_distributions::fprob(($k),($N-($k+1)),($F)));
  $r->{F}                   = $F;
  $r->{F_prob}              = $F_prob;
  $self->{gb}{last_run}{F}      = $F;
  $self->{gb}{last_run}{F_prob} = $F_prob;


  $i = 1;

  foreach my $ev ( ("_cons", @indeps) ) {
   my $disp   = m_get_elem_s($disp_m,$i,$i);
   my $stderr = sqrt($disp);
   $r->{stderrs}{$ev}               = $stderr;
   $r->{t_scores}{$ev}              = $r->{coeffs}{$ev} / $stderr;
   $self->{gb}{last_run}{stderrs}{$ev}  = $stderr;
   $self->{gb}{last_run}{t_scores}{$ev} = $r->{coeffs}{$ev} / $stderr;
   $i++;
  }

  $r->{status}              = "ok";
  $self->{gb}{last_run}{status} = "ok";
  return $r;
 } else {
  my $ses = m_gen_sqrt_diag_a($disp_m);
  my $i = 0;
  foreach my $ev ( ("_cons", @indeps) ) {
   my $stderr = $ses->[$i];
   $r->{stderrs}{$ev} = $stderr;
   $self->{gb}{last_run}{stderrs}{$ev}  = $stderr;
   $r->{t_scores}{$ev}              = $r->{coeffs}{$ev} / $stderr;
   $i++;
  }

  my $R_width = @indeps + 1;
  my $R_depth = @indeps;
#  print "Rwidth $R_width Rdepth $R_depth\n";
  my $R_m = gen_zero_matrix_m($R_depth,$R_width);
  for (my $p=0;$p<@indeps;$p++) {
   m_set_elem($R_m, $p+1,$p+2,1.0);
  }
  my $r_m = gen_zero_matrix_m($R_depth,1);
  my $tr = calc_ols_WaldF($self->{gb},$R_m,$r_m);
  if (func_fail($tr)) { 
   return $tr;
  } 
  my $F = $tr->{F};
  my $F_prob = $tr->{Prob_F};

  $r->{F}                   = $F;
  $self->{gb}{last_run}{F}      = $F;
  #my $F_prob = abs(Statistics::Distributions::fprob($k,$N-($k+1),$F));
  $r->{F_prob}              = $F_prob;
  $self->{gb}{last_run}{F_prob} = $F_prob;
  $r->{status}              = "ok";
  $self->{gb}{last_run}{status} = "ok";
  return $r;
 }
}

sub reg {
 my $self      = shift;
 my $a         = shift;
 return $self->regress($a);
}

sub regress {
 my $self      = shift;
 my $a         = shift;
 my $argref    = $a->{args};
 my $oo        = $self->{oo};

 my $r = $self->regress_calc($a,{ instrumental_vars => 0});
 if (func_fail($r)) { return $r; };
 regress_pretty_print($self->{gb},$self->{oo},$a,$r);
 $r->{status} = "ok";
 return $r;
}

sub regress_pretty_print {
 my $gb   = shift;
 my $oo   = shift;
 my $a    = shift;
 my $r    = shift;

 ### Pretty Print
 my $os = "";
 $os .= "\n\n";
 if (tstFlg($a,"robust")) {
  $os .=
  sprintf(" Regression with Robust Standard Errors (HC1)    Number of obs = %d\n",$r->{N}) .
  sprintf("                                                 F(%3d,%5d)  = %4.2f\n",$r->{k},$r->{N}-($r->{k}+1),$r->{F}) .
  sprintf("                                                 Prob > F      = %1.3f\n",$r->{F_prob});
  $os .=
  sprintf("                                                 R-squared     = %f\n",$r->{rsquared}) .
  sprintf("                                                 Root MSE      = %s\n",fixedFloat_p($r->{sigma_e})) .
  sprintf("\n");
  $os .=
  sprintf("-----------------------------------------------------------------------------\n") .
  sprintf(" %10.10s |      Coef.   Std. Err.      t    P>|t|     [%2d%% Conf. Interval]\n",$r->{depvar},100*$gb->{conf}) .
  sprintf("------------+---------------------------------------------------------------\n");
 } else {
  $os .= 
  sprintf("     Source |       SS       df       MS         Number of obs = %d\n",$r->{N}) .
  sprintf("------------+------------------------------      F(%3d,%5d)  = %4.2f\n",$r->{k},$r->{N}-($r->{k}+1),$r->{F}) .
  sprintf("      Model | %s %5d %s        Prob > F      = %1.3f\n",fixedFloat($r->{ess}),$r->{k},fixedFloat($r->{mse}),$r->{F_prob});
  $os .=
  sprintf("   Residual | %s %5d %s        R-squared     = %f\n",fixedFloat($r->{rss}),$r->{N}-($r->{k}+1),fixedFloat($r->{msr}),$r->{rsquared}) .
  sprintf("------------+------------------------------      Adj R-squared = %s\n",fixedFloat_p($r->{adj_rsquared})) . 
  sprintf("      Total | %s %5d %s        Root MSE      = %s\n",fixedFloat($r->{tss}),$r->{N}-1,fixedFloat($r->{mst}), fixedFloat_p($r->{sigma_e})) .
  sprintf("\n");
  $os .=
  sprintf("-----------------------------------------------------------------------------\n") .
  sprintf(" %10.10s |      Coef.   Std. Err.      t    P>|t|     [%2d%% Conf. Interval]\n",$r->{depvar},100*$gb->{conf}) .
  sprintf("------------+---------------------------------------------------------------\n");
 }

 foreach my $ev ( sort keys %{$r->{coeffs}} ) {
  my $t_dof = $r->{N} - $r->{k} + 1;
  my $t_prob = mod_distributions::tprob($t_dof,abs($r->{t_scores}{$ev}));
#  print "TDOF $t_dof TPROB $t_prob T_VAL $r->{t_scores}{$ev}\n";
  $t_prob *= 2;
  my $ev_copy = $ev;
  if (length($ev_copy) > 10) {
   while(length($ev_copy) > 10) {
    $os .= sprintf(" %10.10s |\n",(" " . substr($ev_copy,0,8,'') . "-"));
   }
   $os .= sprintf(" %10.10s | ",$ev_copy);
  } else {
   $os .= sprintf(" %10.10s | ",$ev);
  }
  $os .= sprintf("%s %s %s %1.5f %s %s\n",
	  fixedFloat($r->{coeffs}{$ev}), 
	  fixedFloat_p($r->{stderrs}{$ev}), 
	  fixedFloat($r->{t_scores}{$ev}),
	  $t_prob,
          fixedFloat($r->{coeffs}{$ev} - $r->{stderrs}{$ev}*$r->{t_crit}),
	  fixedFloat($r->{coeffs}{$ev} + $r->{stderrs}{$ev}*$r->{t_crit}));
  
 };
 $os .= "\n\n";
 $oo->disp($os);
}



sub exclusionrestrict {
 my $self      = shift;
 my $a         = shift;
 my $oo        = $self->{oo}; 

 my $r = {};
 # print Dumper($self->{gb}{last_run},$self->{gb}{last_last_run});

 my $rsquared_ur = $self->{gb}{last_run}{rsquared};
 my $dv_ur       = $self->{gb}{last_run}{depvar};
 my @evs_ur      = @{$self->{gb}{last_run}{indeps}};
 my $k_ur        = $self->{gb}{last_run}{k};
 my $N_ur        = $self->{gb}{last_run}{N};
 my $rsquared_r  = $self->{gb}{last_last_run}{rsquared};
 my $dv_r        = $self->{gb}{last_last_run}{depvar};
 my @evs_r       = @{$self->{gb}{last_last_run}{indeps}};
 my $k_r         = $self->{gb}{last_last_run}{k};
 my $N_r         = $self->{gb}{last_last_run}{N};
 my $l_type      = $self->{gb}{last_run}{type};
 my $ll_type     = $self->{gb}{last_last_run}{type};

 if (!defined($rsquared_ur) ||
     !defined($k_ur) ||
     !defined($N_ur) ||
     !defined($rsquared_r) ||
     !defined($k_r) ||
     !defined($N_r)) {
  return { status => "fail; the last two commands must be regressions" }
 }
 if ($N_ur != $N_r) {
  return { status => "fail; the last two regressions were on different size samples" }
 }
 if ($dv_ur ne $dv_r) {
  return { status => "fail; the last two regressions were on different dependent vars" }
 }
 
# if (($l_type eq "robust") || 
#     ($ll_type eq "robust")) {
#  return { status => "fail; sorry exclusionrestrict doesn't (yet) work with HC robust regressions" };
# }

 if ((($l_type eq "robust") && ($ll_type ne "robust")) ||
     (($l_type ne "robust") && ($ll_type eq "robust"))) {
  print "L_TYPE is $l_type LL_TYPE is $ll_type\n";
  return { status => "fail; can't have one regression \"robust\" and the other not" };
 }

 if (!(($self->{gb}{last_last_run}{status} eq "ok") && ($self->{gb}{last_run}{status} eq "ok"))) {
  return { status => "fail; both regressions did not complete" }
 }

 if ($k_ur < $k_r) {
  my $tk       = $k_ur;
  my $tn       = $N_ur;
  my $trs      = $rsquared_ur;
  my $tdv      = $dv_ur;
  my @tevs     = @evs_ur;
  $k_ur        = $k_r;
  $N_ur        = $N_r;
  $rsquared_ur = $rsquared_r;
  $dv_ur       = $dv_r;
  @evs_ur      = @evs_r;
  $k_r         = $tk;
  $N_r         = $tn;
  $rsquared_r  = $trs;
  $dv_r        = $tdv;
  @evs_r       = @tevs;
}

 foreach my $ev (@evs_r) {
  if (!inAry($ev,\@evs_ur)) {
   return { status => "fail; unrestricted model not perfect superset of restricted" }
  }
 }

 $r->{N}           = $N_ur;
 $r->{k_ur}        = $k_ur;
 $r->{k_r}         = $k_r;
 $r->{rsquared_ur} = $rsquared_ur;
 $r->{rsquared_r}  = $rsquared_r;

 # print Dumper($self->{gb});

 my $q_ur = $k_ur - $k_r;

 my $dof_d = $N_ur - $k_ur - 1;
 my $dof_n = $q_ur;

 my $F_n = (($rsquared_ur - $rsquared_r)/$dof_n);
 my $F_d = (1-$rsquared_ur)/($dof_d);
 my $F = $F_n / $F_d;
 $r->{dof_n} = $dof_n;
 $r->{dof_d} = $dof_d;
 $r->{q}     = $q_ur;
 $r->{F}     = $F;

 $r->{status} = "ok";
 my $Fprob = mod_distributions::fprob($dof_n,$dof_d,$F);
 $r->{Fprob} = $Fprob;

 my $os = "\n\n";
 $os .= " Regression Significance Comparison Test\n";
 $os .= " ------------------------------------------------------------------\n";
 $os .= "                    Unrestricted    Restricted     Comparison\n";
 $os .= "\n";
 $os .= sprintf(" dependent var     %13.13s %13.13s\n",$dv_ur, $dv_r); 
 $os .= "\n";
 foreach my $ev (sort @evs_ur) {
  my $rev = (inAry($ev,\@evs_r)) ? $ev : "";
  $os .= sprintf(" indep var         %13.13s %13.13s\n",$ev, $rev); 
 }
 $os .= "\n";
 $os .= sprintf(" r-squared            %s    %s\n",fixedFloat($rsquared_ur), fixedFloat($rsquared_r));
 $os .= "\n";
 $os .= sprintf(" F_dof_numerator                                   %10d\n",$dof_n);
 $os .= sprintf(" F_dof_denominator                                 %10d\n",$dof_d);
 $os .= sprintf(" F                                                 %s\n",fixedFloat($F));
 $os .= sprintf(" Prob(F)                                           %s\n",fixedFloat($Fprob));
 $os .= "\n\n";
 $oo->disp($os);
 return $r; 
 
}



sub test {
 my $self      = shift;
 my $a         = shift;

 my @argary    = @{$a->{args}};


 my $r = {};

 my @vars;
 my $state = "nothing";
 foreach my $v (@argary) {
  if ($state eq "nothing") {
   if (defined(($self->{gb}{last_run}{coeffs}{$v}))) {
     push (@vars, $v);
     $state = "have_1st_var";
   } else {
    return { status => "fail; variable $v not in regression" }
   }
  } elsif ($state eq "have_1st_var") {
   if (defined($self->{gb}{last_run}{coeffs}{$v})) {
    push (@vars, $v);
    $state = "n_vars_ne_zero";
   } elsif ($v eq "=") {
    $state = "first_var_have_eq";
   } else {
    return { status => "fail; not understood $v" }
   }
  } elsif ($state eq "several_vars_ne_zero") {
   if (defined($self->{gb}{last_run}{coeffs}{$v})) {
    push(@vars,$v);
    $state = "n_vars_ne_zero";
   } else {
    return { status => "fail; variable $v not in regression" }
   }
  } elsif ($state eq "first_var_have_eq") {
   if (defined($self->{gb}{last_run}{coeffs}{$v})) {
    push(@vars, $v);
    $state = "n_var_eq";
   } else {
    my @e = @argary;
    shift @e; shift @e;
    my $eqn = "";
    while (@e) {
     $eqn .= shift @e;
    }
    $r = create_evaluable_eqn_immed($self->{dm},$self->{gb},$eqn);
    if (func_fail($r)) { return $r; }
    goto done_w_args;
   }
  } elsif ($state eq "n_var_eq") {
   if (defined($self->{gb}{last_run}{coeffs}{$v})) {
    push(@vars, $v);
   } elsif ($v eq "=") {
   } else {
    return { status => "fail; garbled command" }
   }
  } elsif ($state eq "n_vars_ne_zero") {
   if (defined($self->{gb}{last_run}{coeffs}{$v})) {
    push(@vars, $v);
   } else {
    return { status => "fail; var $v not in regression" }
   }
  }
 }
 done_w_args:
 $r->{vars} = \@vars;
 $r->{state} = $state;
 $r->{status} = "ok";
 my @indeps = ( "_cons" );
 push(@indeps, @{$self->{gb}{last_run}{indeps}});

 my $R_m;
 my $r_m;
 my $evd;
 my $coeff_aa = [];
 if ($state eq "n_var_eq") {
  for (my $i=0;$i<@indeps;$i++) {
   $evd->{$indeps[$i]} = $i;
  }
  for (my $i=0;$i<(@vars-1);$i++) {
   my $v1 = $vars[$i];
   my $v2 = $vars[$i+1];
   for (my $j=0;$j<@indeps;$j++) {
    $coeff_aa->[$i][$j] = 0;
   }
   $coeff_aa->[$i][$evd->{$v1}] = 1;
   $coeff_aa->[$i][$evd->{$v2}] = -1;
  }
  $R_m = a_gen_matrix_from_rows_m($coeff_aa);
  $r_m = gen_zero_matrix_m(@vars-1,1);
 } elsif ($state eq "n_vars_ne_zero") {
  for (my $i=0;$i<@indeps;$i++) {
   $evd->{$indeps[$i]} = $i;
  }
  for (my $i=0;$i<(@vars);$i++) {
   my $v1 = $vars[$i];
   for (my $j=0;$j<@indeps;$j++) {
    $coeff_aa->[$i][$j] = 0;
   }
   $coeff_aa->[$i][$evd->{$v1}] = 1;
  }
  my $nvars = @vars;
  $R_m = a_gen_matrix_from_rows_m($coeff_aa);
#  $r_m = gen_zero_matrix_m($#vars+1,1); 
  $r_m = gen_zero_matrix_m($nvars,1); 
 } else {
  return { status => "fail; all test can handle right now is testing equality among betas" }
 }

 $r = calc_ols_WaldF($self->{gb},$R_m,$r_m);
 if (func_fail($r)) { return $r; }
 my $F = $r->{F};
 my $Pr_F = $r->{Prob_F};

 my $os = "\n\n";
 $os .= " Test of regression coefficients...\n\n";

 my ($rs, $cs) = m_ret_size_a($R_m);
 for (my $i=1;$i<=$rs;$i++) {
  $os .= "   ( $i)   ";
  my @row = ();
  for (my $j=1;$j<($cs);$j++) {
   $row[$j] = m_get_elem_s($R_m,$i,$j+1);
   if ($row[$j]) {
    $os .= " " . $row[$j] .' '. $self->{gb}{last_run}{indeps}[$j-1] . " +";
   }     
  }
  chop $os;
  $os .= " = " . m_get_elem_s($r_m,$i,1);
  $os .= "\n";
 }
 $os .= "\n";
 $os .= sprintf("  F(%d,%d) = %f\n",$r->{dof_n},$r->{dof_d},$F);

 $os .= sprintf("  Prob > F = %f\n",$Pr_F);
 $os .= "\n\n";

 my $oo = $self->{oo};
 $oo->disp($os);
 # print Dumper($r);
 return $r;
}


sub vce {
 my $self = shift;
 my $a    = shift;
 my $oo   = $self->{oo};

 my $disp_m = $self->{gb}{last_run}{matrices}{disp_m};
 if (defined($disp_m)) {
  my $os = "\nCovariance matrix of coefficients of regression model\n\n";
  my @vlist = ("_const", @{$self->{gb}{last_run}{indeps}});

  $os .= "           |";
  foreach my $evc (@vlist) {
   $os .= sprintf("%10.10s ",$evc);
  }
  $os .= "\n";
  $os .= "-----------+";
  foreach my $evc (@vlist) {
   $os .= sprintf("-----------");
  }
  $os .= "\n";
  my $c=1;
  foreach my $evr (@vlist) {
   my $r=1;
   $os .= sprintf("%10.10s | ",$evr);
   foreach my $evc (@vlist) {
    $os .= sprintf("%s ", fixedFloat(m_get_elem_s($disp_m,$c,$r)));
    $r++;
   }
   $os .= "\n";
   $c++;
  }
  $os .= "\n\n";
  $oo->disp($os);
  return { status => "ok" }
 } else {
  return { status => "fail; no covariance matrix in memory" }
 }
}



sub calc_ols_WaldF {

 my $gb   = shift;
 my $R_m  = shift;
 my $r_m  = shift;

 my $b_m = $gb->{last_run}{matrices}{soln_m};
 my $disp_m = $gb->{last_run}{matrices}{disp_m};
 my ($wald_chi, $R_rows);
 ($wald_chi, $R_rows) = m_calc_ols_wald_s($R_m,$r_m,$b_m,$disp_m);
 my $dof_n = $R_rows; 
 my $dof_d = $gb->{last_run}{N} - $gb->{last_run}{k} -1;

 my $F = $wald_chi / $dof_n;
 my $Pr_F = mod_distributions::fprob($dof_n,$dof_d,$F);
 my $r = {};
 $r->{matrix_result} = $wald_chi; 
 $r->{F}             = de_objify($F);
 $r->{Prob_F}        = $Pr_F;
 $r->{status}        = "ok";
 $r->{dof_n}         = de_objify($dof_n);
 $r->{dof_d}         = de_objify($dof_d);
 $r->{coeff_matrix}  = m_mat_to_string($R_m);
 $r->{const_vector}  = m_mat_to_string($r_m);
 return $r;

};

1;

