
#   Copyright (C) 2006 David G. Jacobowitz 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#


package matrix_help;
use Exporter 'import';
@EXPORT_OK = qw(m_col_to_diag m_sq_elems m_gen_tp m_gen_sqrt_diag_a a_scale_const m_return_col_a m_mean_tot_col_s a_ss_s m_ss_col_s a_gen_ifqualed_a a_gen_matrix_from_cols_m a_gen_matrix_from_rows_m gen_zero_matrix_m m_get_elem_s de_objify m_mat_to_string m_ret_size_a set_matrix_mem m_return_rank_s m_sum_sq_s m_print_mat_str m_set_elem m_mul_m_diag_m);

use strict;
use matlibctrl qw(use_pari use_pdl);
use misc qw(deep_copy);
use Data::Dumper;

my $using_pari = use_pari();
my $using_pdl  = use_pdl();

 if      ($using_pdl) {
  use PDL;
  $PDL::BIGPDL = 1;
  use Math::MatrixReal;
 } else {
  use Math::MatrixReal;
 }


sub set_matrix_mem {
 my $newval = shift;
 if ($using_pari) {
  if ($newval =~ /[mM]$/) {
   $newval =~ s/[mM]$//;
   $newval *= 1e6;
  } elsif ($newval =~ /[gG]$/) {
   $newval =~ s/[gG]$//;
   $newval *= 1e9
  }
  allocatemem($newval);
 }
}

sub de_objify {
 my $in = shift;
 if ($using_pari) {
  return pari2num($in);
 } else {
  return $in;
 }
}

sub gen_zero_matrix_m {
 my $rows = shift;
 my $cols = shift;
# print "zero matrix $rows rows $cols cols\n";
 if ($using_pdl) {
  my $o = zeroes($cols,$rows);
  return $o;
 } elsif ($using_pari) {
  my $o = matrix($rows,$cols);
  return $o;
 } else {
  my $o = Math::MatrixReal->new($rows,$cols);
  $o->zero;
  return $o;
 }
}



sub m_ret_size_a {
 my $m = shift;
 my $rs, my $cs;
 if      ($using_pdl) {
  ($cs, $rs) = $m->dims();
 } elsif ($using_pari) {
  my $sz = matsize($m);
  $rs = $sz->[0];
  $cs = $sz->[1];
 } else {
  ($rs, $cs) = $m->dim;
 }
 return ($rs, $cs);
}

sub m_mat_to_string {
 my $m = shift;
 if ($using_pdl) {
  return sprintf("%s",$m);
 } elsif ($using_pari) {
  return pari2pv($m);
 } else {
  return $m->as_yacas();
 }
}

sub m_set_elem {
 my $m = shift;
 my $row = shift;
 my $col = shift;
 my $v   = shift;
 if ($using_pdl) {
  $m->set($col-1,$row-1,$v);
 } elsif ($using_pari) {
  $m->[$col-1][$row-1] = $v;
 } else {
  $m->assign($row,$col,$v);
 }
 return $m;
}


sub m_get_elem_s {
 my $m   = shift;
 my $row = shift;
 my $col = shift;
 if ($using_pdl) {
  return $m->at($col-1,$row-1);
 } elsif ($using_pari) {
  return $m->[$col-1][$row-1];
 } else {
  return $m->element($row,$col);
 }
}

sub a_gen_matrix_from_cols_m {
 my $cols_a = shift;
 if ($using_pdl) {
  my $mt = double(pdl($cols_a));
  my $m  = $mt->transpose();
  return $m;
 } elsif ($using_pari) {
  ## this is a heinous hack because there is something deeply wrong
  ## with the pari lib and it turns some data into integers 
  ## randomly. I have to scan the matrix myself and make sure
  ## that pari knows these are floats by dividing by 1.
  #my $cols = @{$cols_a};
  #my $rows = @{$cols_a->[0]};
  #print "cols $cols rows $rows\n";
  #my $m = matrix($rows,$cols);
  #for (my $c=0;$c<$cols;$c++) {
  # for (my $r=0;$r<$rows;$r++) {
  #  print "col $c row $r\n";
  #  $m->[$c][$r] = de_objify($cols_a->[$c][$r]/1);
  # }
  #}
  my $m = PARImat($cols_a);
  return $m;
 } else {
  my $m = Math::MatrixReal->new_from_cols($cols_a);
 return $m;
 }
}

sub a_gen_matrix_from_rows_m {
 my $rows_a = shift;
 if ($using_pdl) {
  my $m = double(pdl($rows_a));
 } elsif ($using_pari) {
  my $m = PARImat_tr($rows_a);
  return $m;
 } else {
  my $m = Math::MatrixReal->new_from_rows($rows_a);
 return $m;
 }
}


# take two arrays that should be of equal length and generate
# a third array  that consists only of data from the first array
# at whose index the second array is nonzero. This is for the 'if'
# clause in many functions
sub a_gen_ifqualed_a {
 my $da = shift;
 my $ia = shift;
 my $oa = [];
 my $r = 0;
 for (my $q=0;$q<@{$da};$q++) {
  if ($ia->[$q]) { $oa->[$r++] = $da->[$q]; };
 }
 return $oa;
}

# convert a single column matrix into a square matrix
# where each row of the old is placed into the diagonals of the 
# new
sub m_col_to_diag {
 my $cm = shift;
 if ($using_pdl) {
  # there must be a more efficient way to do this, but I can't
  # figure out the PDL docs right now.
  my ($cs,$rs) = $cm->dims();
  my $dm = zeroes($rs,$rs);
  for (my $i=0;$i<$rs;$i++) {
   $dm->set($i,$i,$cm->at(0,$i));
  }
  return $dm;
 } elsif ($using_pari) {
  my $dm = matdiagonal( $cm->[0] );
  return $dm;
 } else {
  my ($rows, my $cols) = $cm->dim();
  if ($cols != 1) { return undef; }
  my $rm = Math::MatrixReal->new($rows,$rows);
  for (my $i=1;$i<$rows+1;$i++) {
   $rm->assign($i,$i,$cm->element($i,1));
  }
  return $rm;
 }
}


# square every element in a matrix, return a copy of
# that result
sub m_sq_elems {
 my $i = shift;
 if ($using_pdl) {
  my $o = pdl $i;
  $o *= $o; 
  return $o;
 } elsif ($using_pari) {
  my $sz = matsize($i);
  my $rows = $sz->[0];
  my $cols = $sz->[1];
  my $o = matrix($rows,$cols);
  for (my $r=0;$r<$rows;$r++) {
   for (my $c=0;$c<$cols;$c++) {
    $o->[$c][$r] = ($i->[$c][$r])**2;
   }
  }
  return $o;
 } else {
  my $o = $i->each( sub { my $e = shift;
		          my $r = shift;
			  my $c = shift;
			  my $e2 = $e*$e;
                          return $e2;
                        });
  return $o;
 }
}

# return a matrix consisting of the transpose of another matrix
sub m_gen_tp {
 my $i = shift;
 if ($using_pdl) {
  my $o = pdl $i;
  return ~$o;
 } elsif ($using_pari) {
  my $o = mattranspose($i);
  return $o;
 } else {
  my ($rs, $cs) = $i->dim();
  my $o = Math::MatrixReal->new($cs,$rs);
  $o->transpose($i);
  return $o;
 }
}

# return an array that contains the sqrt of all the diagonal
# elements in the matrix
sub m_gen_sqrt_diag_a {
 my $i = shift;
 if ($using_pdl) {
  my ($cols,$rows) = $i->dims();
  if ($rows == $cols) {
   my $o = [];
   for (my $c=0;$c<$cols;$c++) {
    $o->[$c] = sqrt($i->at($c,$c));
   }
   return $o;
  } else {
   return undef;
  }
 } elsif ($using_pari) {
  my $sz = matsize($i);
  my $rows = $sz->[0];
  my $cols = $sz->[1];
  if ($rows == $cols) {
   my $o = [];
   for (my $c=0;$c<$rows;$c++) {
    $o->[$c] = sqrt($i->[$c][$c]);
   }
   return $o;
  } else {
   return undef;
  }
 } else {
  my ($rs, $cs) = $i->dim;
  if ($rs == $cs) {
   my $o = [];
   for (my $c=1;$c<$rs+1;$c++) {
    $o->[$c-1] = sqrt($i->element($c,$c));
   }
   return $o;
  } else {
   return undef;
  }
 }
}

# return all the diagonal elements in an matrix as an array;
sub m_gen_diag_a {
 my $i = shift;
 if ($using_pdl) {
  my $o = [];
  my ($cols,$rows) = $i->dims();
  if ($rows == $cols) {
   for (my $c=0;$c<$rows;$c++) {
    $o->[$c] = $i->at($c,$c);
   }
   return $o;
  } else {
   return undef;
  }
 } elsif ($using_pari) {
  my $o = [];
  my $sz = matsize($i);
  my $rows = $sz->[0];
  my $cols = $sz->[1];
  if ($rows == $cols) {
   for (my $c=0;$c<$rows;$c++) {
    $o->[$c] = $i->[$c][$c];
   }
  } else {
   return undef;
  } 
 } else {
  my ($rs, $cs) = $i->dim;
  if ($rs == $cs) {
   my $o = [];
   for (my $c=1;$c<$rs+1;$c++) {
    $o->[$c-1] = $i->element($c,$c);
   }
   return $o;
  } else {
   return undef;
  }
 }
}


sub m_mean_tot_col_s {
 my $im = shift;
 my $c  = shift;
 if ($using_pdl) {
  my ($cols,$rows) = $im->dims();
  if (!defined($im) || ($cols < $c)) {
   return undef;
  }
  my $t = 0;
  for (my $r=0;$r<$rows;$r++) {
   $t += $im->at($c-1,$r);
  }
  return ($t/$rows, $t);
 } elsif ($using_pari) {
  my $sz = matsize($im);
  my $rows = $sz->[0];
  my $cols = $sz->[1];
  if (!defined($im) || ($cols < $c)) {
   return undef;
  }
  my $t = 0;
  for (my $r=0;$r<$rows;$r++) {
   $t += $im->[$c-1][$r];
  }
  return ($t/$rows, $t);
 } else {
  my ($rs, $cs) = $im->dim;
  if (!defined($im) || ($cs < $c)) {
   return undef;
  } 
  my $t = 0;
  for (my $r=0;$r<$rs;$r++) {
   $t += $im->element($r+1,$c);
  }
  return ($t/$rs, $t);
  }
}

# scale an array by a constant
#
sub a_scale_const {
 my $ia = shift;
 my $ic = shift;
 my $oa = [];
 foreach (my $q=0;$q<@{$ia};$q++) {
  $oa->[$q] = $ia->[$q] * $ic;
 }
 return $oa;
}

# return a column of a matrix as an array
sub m_return_col_a {
 my $im = shift;
 my $c  = shift;
 if ($using_pdl) {
  my ($cs, $rs) = $im->dims();
  if (!defined($im) || ($cs < $c)) {
   return undef;
  }
  my $oa = [];
  for (my $r=0;$r<$rs;$r++) {
   $oa->[$r] = $im->at($c-1,$r);
  }
  return $oa;
 } elsif ($using_pari) {
  my $sz = matsize($im);
  my $rs = $sz->[0];
  my $cs = $sz->[1];
  if (!defined($im) || ($cs < $c)) {
   return undef;
  } 
  my $oa = [];
  for (my $r=0;$r<$rs;$r++) {
   $oa->[$r] = pari2num($im->[$c-1][$r]);
  }
  return $oa;
 } else {
  my ($rs, $cs) = $im->dim;
  if (!defined($im) || ($cs < $c)) {
   return undef;
  }
  my $oa = [];
  for (my $r=0;$r<$rs;$r++) {
   $oa->[$r] = $im->element($r+1,$c);
  }
  return $oa;
 }
}


# return the sum of squares of the difference of each element of a column
# of a matrix and a supplied scalar
sub m_ss_col_s {
 my $im = shift;
 my $m  = shift;
 my $c  = shift;
 if ($using_pdl) {
  my $cim = pdl $im;
  $cim -= $m;
  $cim *= $cim;
  my $ss = $cim->sum();
  return $ss;
 } elsif ($using_pari) {
  my $sz = matsize($im);
  my $rs = $sz->[0];
  my $cs = $sz->[1];
  if (!defined($im) || ($cs < $c)) {
   return undef;
  }
  my $ss = 0;
  for (my $r=0;$r<$rs;$r++) {
   $ss += ($im->[$c-1][$r] - $m) *
          ($im->[$c-1][$r] - $m);
  }
  return $ss;
 } else {
  my ($rs, $cs) = $im->dim;
  if (!defined($im) || ($cs < $c)) {
   return undef;
  }
  my $ss = 0;
  for (my $r=0;$r<$rs;$r++) {
   $ss += ($im->element($r+1,$c) - $m)*
          ($im->element($r+1,$c) - $m);
  }
  return $ss;
 }
}


# return the sum of squares of difference of array values with some supplied
# scalar
sub a_ss_s {
 my $a = shift;
 my $m = shift;
 my $ss = 0;
 for (my $i=0;$i<@{$a};$i++) {
  $ss += ($a->[$i] - $m)*
         ($a->[$i] - $m);
 }
 return $ss;
}

sub m_return_rank_s {
 my $m = shift;
 my $r = 0;
 my $i = 0;
 if ($using_pdl) {
  # PDL doesn't have a rank function, though PDL::LinearAlgebra
  # does. Unfortunately, there is no available binary for that 
  # package for Win32 (or Mac for that matter) so we'll just 
  # suck it up and use MatrixReal
 } elsif ($using_pari) {
  my $sz = matsize($m);
  my $rs = $sz->[0];
  my $cs = $sz->[1];
  if ($rs = $cs) {
   $r = matrank($m);
   if ($r == $cs) { $i = 1; };
  }
 } else {
  my ($rs, $cs) = $m->dim();
  if ($rs == $cs) {
   my $lr = $m->decompose_LR();
   $r = $lr->rank_LR();
   if ($r == $cs)  { $i = 1;}
  }
 }
 return ($r, $i);
}

sub m_sum_sq_s {
 my $m = shift;
 if ($using_pdl) {
  my $t = ~$m;
  my $r = $t x $m;
  return $r->at(0,0);
 } elsif ($using_pari) {
  my $t = mattranspose($m);
  my $r = $t * $m;
  return $r->[0][0];
 } else {
  my ($rs, $cs) = $m->dim();
  my $t = Math::MatrixReal->new($cs,$rs);
  $t->transpose($m);
  my $r = $t * $m;
  return $r->element(1,1);
 }
}

sub m_print_mat_str {
 my $m = shift;
 my $os = "";
 if ($using_pdl) {
  my ($cols,$rows) = $m->dims();
  $os .= "matrix: (ROWS $rows COLS $cols)\n";
  $os .= "[";
  for (my $x=0;$x<$rows;$x++) {
   $os .= "[\t";
   for (my $y=0;$y<$cols;$y++) {
    $os .= sprintf("%5.5f\t",$m->at($y,$x));
   }
   chop $os;
   $os .= "]\n";
  }
 } elsif ($using_pari) {
  my $sz = matsize($m);
  my $rows = $sz->[0];
  my $cols = $sz->[1];
  $os .= "matrix: (ROWS $rows COLS $cols)\n";
  $os .= "[";
  for (my $x=0;$x<$rows;$x++) {
   $os .= "[\t";
   for (my $y=0;$y<$cols;$y++) {
    $os .= sprintf("%5.5f\t",$m->[$y][$x]);
   }
   chop $os;
   $os .= "]\n";
  }
  chop $os;
  $os .= "]\n";
 } else {
  my ($rows, $cols) = $m->dim();
  $os .= "[";
  for (my $y=1;$y<=$cols;$y++) {
   $os .= "[\t";
   for (my $x=1;$x<=$rows;$x++) {
    $os .= sprintf("%5.5f\t",$m->element($x,$y));
   }
   chop $os;
   $os .= "]\n";
  }
  chop $os;
  $os .= "]\n";
 }
 return $os;
}


# this subroutines takes a matrix M and multiplies it by
# the diagonal matrix D that would be formed from a  
# column vector D_v. If that column vector were really long, 
# its matrix would be big, and mostly empty. We can be much
# more efficient in time and space knowing that only the diagonals
# are nonzero...
sub m_mul_m_diag_m {
 my $M  = shift;
 my $Dv = shift;
 if ($using_pdl) {
  my ($M_cs,$M_rs) = $M->dims();
  my ($D_cs,$D_rs) = $Dv->dims();
  if ($M_cs != $D_rs) {
   return undef;
  }
  my $R = zeroes($M_cs,$M_rs);
  for (my $c=0;$c<$D_rs;$c++) {
   for (my $r=0;$r<$M_rs;$r++) {
    my $dv = $Dv->at(0,$c);
    my $mv = $M->at($c,$r);
    my $pv = $dv * $mv;
    $R->set($c,$r,$pv);
   }
  }
  return $R;
 } elsif ($using_pari) {
  return matmuldiagonal($M,$Dv);
 } else {
  my ($D_rs, $D_cs) = $Dv->dim();
  my ($M_rs, $M_cs) = $M->dim();
  if ($M_cs != $D_rs) {
   return undef;
  }
  my $R = Math::MatrixReal->new($M_rs,$M_cs);
  for (my $c=1;$c<=$D_rs;$c++) {
   for (my $r=1;$r<=$M_rs;$r++) {
    my $dv = $Dv->element($c,1);
    my $mv = $M->element($r,$c);
    my $pv = $dv * $mv;
    $R->assign($r,$c,$pv);
   }
  }
  return $R;
 }
}

1;


